/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd
 *
 * Licensed under the Flora License, Version 1.1 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://floralicense.org/license/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <utils_i18n.h>
#include <stdlib.h>
#include "data.h"
#include "$(appName).h"

/**
 * @brief Function acquires all the available locale using the i18n and reports
 * them to the controller module.
 * Internally, for each locale reported by the i18n the provided controller's
 * callback function is invoked.
 * @param[in] func_cb The callback function invoked separately for each locale
 * acquired from the i18n.
 * @return This function returns 'true' if all the available locale were successfully
 * acquired, otherwise 'false' is returned.
 */
bool data_get_available_locale(get_available_locale_cb func_cb)
{
	int i;
	int32_t count;
	const char *locale;

	if (!func_cb) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Wrong argument provided.");
		return false;
	}

	count = i18n_ulocale_count_available();

	for (i = 0; i < count; i++) {
		locale = i18n_ulocale_get_available(i);

		if (locale && !func_cb(locale))
			break;
	}

	return true;
}

/**
 * @brief Function acquires the language's display name using i18n,
 * based on the provided localization string.
 * @param[in] locale The localization string.
 * @param[out] display_language The language's display name acquired
 * based on the localization string.
 * @return This function returns 'true' if language's display name was successfully
 * acquired, otherwise 'false' is returned.
 */
bool data_get_locale_display_language(const char *locale, char **display_language)
{
	int32_t str_len;
	i18n_uchar i18n_language[256] = {0,};

	if (!display_language) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Wrong argument provided.");
		return false;
	}

	i18n_ulocale_get_display_language(locale, NULL, i18n_language, sizeof(i18n_language));
	int ret = get_last_result();
	if (ret != I18N_ERROR_NONE) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Function i18n_ulocale_get_display_language() failed with error %d.", ret);
		return false;
	}

	str_len = i18n_ustring_get_length(i18n_language);
	*display_language = (char *)calloc(str_len + 1, sizeof(char));
	if (!(*display_language)) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Function calloc() failed.");
		return false;
	}

	i18n_ustring_copy_au_n(*display_language, i18n_language, str_len);
	ret = get_last_result();
	if (ret != I18N_ERROR_NONE) {
		free(*display_language);
		*display_language = NULL;
		dlog_print(DLOG_ERROR, LOG_TAG, "Function i18n_ustring_copy_au_n() failed with error %d.", ret);
		return false;
	}

	return true;
}
