/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd
 *
 * Licensed under the Flora License, Version 1.1 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://floralicense.org/license/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <Eina.h>
#include <efl_extension.h>
#include <utils_i18n.h>
#include <glib.h>
#include <math.h>
#include "$(appName).h"
#include "view.h"
#include "view_defines.h"

#define INDICATOR_H 52
#define TITLE_LENGTH 64
#define MAX_INDICATOR_PAGE_NO 4
#define LOCALE_LEN 16
#define SWIPE_RIGHT_ANGLE_LO 75.0
#define SWIPE_RIGHT_ANGLE_HI 105.0
#define SWIPE_LEFT_ANGLE_LO 255.0
#define SWIPE_LEFT_ANGLE_HI 285.0

#define ITEM_INFO "item_info"
#define FOLDER_FLAG_LIST "flag_list"
#define FOLDER_FLAG_CURRENT "flag_current"

typedef enum { INDICATOR_OFF = 0, INDICATOR_ON } indicator_state;

typedef struct _data_pass {
	Evas_Object *parent;
	const char *locale;
} data_pass_t;

static struct view_info {
	Evas_Object *win;
	Evas_Object *conform;
	Evas_Object *layout_main;
	Evas_Object *layout_language;
	Evas_Object *layout_available_langs;
	Elm_Gengrid_Item_Class *grid_itc;
	get_display_language_cb get_language_cb;
	Evas_Object *scroller;
	Evas_Object *box;
	int current_page;
	const char *current_lang;
	char new_lang[LOCALE_LEN];
	double drag_start_t;
	double drag_end_t;
	bool anim_in_progress;
	bool lang_changing;
} s_info = {
	.win = NULL,
	.conform = NULL,
	.layout_main = NULL,
	.layout_language = NULL,
	.layout_available_langs = NULL,
	.grid_itc = NULL,
	.get_language_cb = NULL,
	.scroller = NULL,
	.box = NULL,
	.current_page = 0,
	.current_lang = NULL,
	.new_lang = {0,},
	.drag_start_t = 0,
	.drag_end_t = 0,
	.anim_in_progress = false,
	.lang_changing = false,
};

static void _delete_win_request_cb(void *data, Evas_Object *obj, void *event_info);
static void _layout_back_cb(void *data, Evas_Object *obj, void *event_info);
static void _change_language_cb(void *data, Evas_Object *obj, const char *emission, const char *source);
static void _change_global_lang_cb(void *data, Evas_Object *obj, const char *emission, const char *source);
static void _mouse_down_cb(void *data, Evas *e, Evas_Object *obj, void *event_info);
static void _mouse_up_cb(void *data, Evas *e, Evas_Object *obj, void *event_info);
static void _item_on_del_cb(void *data, Evas *e, Evas_Object *obj, void *event_info);
static Evas_Event_Flags _n_finger_line_start_cb(void *data, void *event_info);
static Evas_Event_Flags _n_finger_line_end_cb(void *data, void *event_info);
static void _send_anim_properties(double time_diff, Evas_Object *layout);
static Evas_Object *_get_lang_item(const char* locale);
static bool _check_swipe_direction(double angle, double x1, double y1, double x2, double y2, double r, double a, double b);
static bool _check_scroll_direction(double angle, int *direction);
static bool _set_language(const char *locale);
static Eina_Bool _add_line_gesture_layer(Evas_Object *parent, Evas_Object *layout, const char *locale);
static void _create_title(const char *locale, char *title, int title_size);
static void _get_app_resource(const char *edj_file_in, char *edj_path_out);
static void _get_country_flag_path(const char *locale_in, const char* folder, char *flag_path_out);
static Evas_Object *_create_item(const char *locale, Evas_Object *parent);
static bool _create_append_item(void *data);
static Evas_Object *_create_current_lang_image(const char *locale);
static bool _change_current_lang_image(const char *locale);
static void _turn_on_off_dot_indicator(int id_old, int id_cur);
static void _toggle_indicator_dot(int page_number, int max_page, indicator_state i_s, int *id);
static Eina_Bool _create_layout_main(void);
static Eina_Bool _create_layout_language(void);
static Eina_Bool _create_layout_available_languages(void);
static Eina_Bool _create_layout_button(void);
static void _display_text_translation(void);
static bool _is_same_string(const char *s1, const char *s2);

/**
 * @brief Creates essential objects: window, conformants and layouts.
 * @param[in] user_data The user data.
 * @return The function returns 'EINA_TRUE' if all objects was created succesfully
 * otherwise 'EINA_FALSE' is returned.
 */
Eina_Bool view_create(void *user_data)
{
	/* Create the window */
	s_info.win = view_create_win(PACKAGE);
	if (s_info.win == NULL) {
		dlog_print(DLOG_ERROR, LOG_TAG, "failed to create a window.");
		return EINA_FALSE;
	}

	if (!_create_layout_main())
		return EINA_FALSE;

	if (!_create_layout_language())
		return EINA_FALSE;

	if (!_create_layout_available_languages())
		return EINA_FALSE;

	if (!_create_layout_button())
		return EINA_FALSE;

	evas_object_show(s_info.win);

	return EINA_TRUE;
}

/**
 * @brief Creates a basic window named package
 * @param[in] pkg_name Name of the window
 * @return The function returns window object if it was created successfully,
 * otherwise 'NULL' is returned.
 */
Evas_Object *view_create_win(const char *pkg_name)
{
	Evas_Object *win = NULL;

	/*
	 * Window
	 * Create and initialize elm_win.
	 * elm_win is mandatory to manipulate the window.
	 */
	win = elm_win_util_standard_add(pkg_name, pkg_name);
	elm_win_conformant_set(win, EINA_TRUE);
	elm_win_indicator_mode_set(win, ELM_WIN_INDICATOR_SHOW);
	elm_win_indicator_opacity_set(win, ELM_WIN_INDICATOR_OPAQUE);
	elm_win_autodel_set(win, EINA_TRUE);

	evas_object_smart_callback_add(win, "delete,request", _delete_win_request_cb, NULL);

	return win;
}

/**
 * @brief Creates a conformant object for parent object.
 * @param[in] parent The parent object for conformant object.
 * @return The function returns conformant object if it was created successfully,
 * otherwise 'NULL' is returned.
 */
Evas_Object *view_create_conformant(Evas_Object *parent)
{
	Evas_Object *conform;
	Evas_Object *bg;

	if (!parent) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Wrong input arguments.");
		return NULL;
	}

	conform = elm_conformant_add(parent);
	if (!conform) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Function elm_conformant_add() failed.");
		return NULL;
	}

	evas_object_size_hint_weight_set(conform, EVAS_HINT_EXPAND, EVAS_HINT_EXPAND);
	elm_win_resize_object_add(parent, conform);

	bg = elm_bg_add(conform);
	if (!bg) {
		dlog_print(DLOG_ERROR, LOG_TAG, "elm_bg_add() failed. Cannot add background to the conformant.");
		return NULL;
	}

	elm_bg_color_set(bg, 0, 0, 0);
	evas_object_size_hint_weight_set(bg, EVAS_HINT_EXPAND, EVAS_HINT_EXPAND);
	elm_object_part_content_set(conform, "elm.swallow.indicator_bg", bg);
	evas_object_show(bg);

	evas_object_show(conform);

	return conform;
}

/**
 * @brief Creates a layout object for parent object based on provided EDJE script.
 * @param[in] parent The parent object for layout object.
 * @param[in] edj_file_name The relative path to the layout EDJE script.
 * @param[in] edj_group The name of the group to be loaded from the EDJE script.
 * @return The function returns layout object if it was created successfully,
 * otherwise 'NULL' is returned.
 */
Evas_Object *view_create_layout(Evas_Object *parent, const char *edj_file_name, const char *edj_group)
{
	char edj_path[PATH_MAX] = {0, };
	Evas_Object *layout = NULL;

	if (!parent || !edj_file_name || !edj_group) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Wrong input arguments.");
		return NULL;
	}

	_get_app_resource(edj_file_name, edj_path);

	layout = elm_layout_add(parent);
	if (!layout) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Function elm_layout_add() failed.");
		return NULL;
	}

	if (!elm_layout_file_set(layout, edj_path, edj_group)) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Function elm_layout_file_set() failed.");
		evas_object_del(layout);
		return NULL;
	}

	evas_object_size_hint_weight_set(layout, EVAS_HINT_EXPAND, EVAS_HINT_EXPAND);

	return layout;
}

/**
 * @brief Creates a layout object and sets it to the parent object based on provided EDJE script.
 * @param[in] parent The parent object for layout object.
 * @param[in] edj_file_name The relative path to the layout EDJE script.
 * @param[in] edj_group The name of the group to be loaded from the EDJE script.
 * @param[in] target_part_name The name of the EDJE part where the layout object has to be set.
 * @return The function returns layout object if it was created successfully,
 * otherwise 'NULL' is returned.
 */
Evas_Object *view_create_set_layout(Evas_Object *parent, const char *edj_file_name, const char *edj_group, const char *target_part_name)
{
	Evas_Object *layout = view_create_layout(parent, edj_file_name, edj_group);
	if (!layout)
		return NULL;

	if (!target_part_name)
		elm_object_content_set(parent, layout);
	else
		elm_object_part_content_set(parent, target_part_name, layout);

	return layout;
}

/**
 * @brief Creates an image object with image file set.
 * @param[in] parent The parent object for image object.
 * @param[in] file_path The path to the image file to be set.
 * @param[in] part_name The name of the EDJE part where the image object has to be set.
 * @return The function returns image object if it was created successfully,
 * otherwise 'NULL' is returned.
 */
Evas_Object *view_create_image(Evas_Object *parent, char *file_path, const char *part_name)
{
	Evas_Object *image = elm_image_add(parent);
	if (!image) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Function elm_image_add() failed.");
		return NULL;
	}

	elm_image_smooth_set(image, EINA_TRUE);
	elm_image_aspect_fixed_set(image, EINA_TRUE);
	elm_image_file_set(image, file_path, NULL);

	if (part_name)
		elm_object_part_content_set(parent, part_name, image);

	return image;
}

/**
 * @brief Creates an box object
 * Items are arranged horizontally
 * @param[in] parent The parent object for image object.
 * @return The function returns box object if it was created successfully,
 * otherwise 'NULL' is returned.
 */
Evas_Object *view_create_box(Evas_Object *parent)
{
	Evas_Object *box = NULL;

	if (!parent) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Wrong input arguments.");
		return NULL;
	}

	box = elm_box_add(parent);
	if (!box) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Function elm_box_add() failed.");
		return NULL;
	}

	elm_object_content_set(parent, box);
	elm_box_horizontal_set(box, EINA_TRUE);
	evas_object_size_hint_align_set(box, EVAS_HINT_FILL, EVAS_HINT_FILL);
	evas_object_size_hint_weight_set(box, EVAS_HINT_EXPAND, EVAS_HINT_EXPAND);
	evas_object_show(box);

	return box;
}

/**
 * @brief Creates a scroller object.
 * @param[in] parent The parent object for the scroller object.
 * @param[in] target_part_name The name of the EDJE part where the scroller object has to be set.
 * @return The function returns scroller object if it was created successfully,
 * otherwise 'NULL' is returned.
 */
Evas_Object *view_create_scroller(Evas_Object *parent, const char *target_part_name)
{
	Evas_Object *scroller = NULL;

	if (!parent || !target_part_name) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Wrong input arguments.");
		return NULL;
	}

	scroller = elm_scroller_add(parent);
	if (!scroller) {
		dlog_print(DLOG_ERROR, LOG_TAG, "view_create_scroller() failed.");
		return NULL;
	}

	elm_scroller_page_relative_set(scroller, 1 / AVAILABLE_LANGS_LIST_ITEMS_PER_PAGE, 0.0);
	elm_scroller_page_size_set(scroller, AVAILABLE_LANGS_LIST_ITEM_WIDTH, AVAILABLE_LANGS_LIST_ITEM_HEIGHT);
	elm_scroller_page_scroll_limit_set(scroller, 3, 1);
	elm_scroller_policy_set(scroller, ELM_SCROLLER_POLICY_OFF, ELM_SCROLLER_POLICY_OFF);

	evas_object_size_hint_weight_set(scroller, EVAS_HINT_EXPAND, EVAS_HINT_EXPAND);
	elm_object_part_content_set(parent, target_part_name, scroller);
	evas_object_show(scroller);

	return scroller;
}

/**
 * @brief Adds callback function invoked on display language name acquisition.
 * @param[in] get_language_cb The callback function invoked on locale selection.
 */
void view_set_callback(get_display_language_cb get_language_cb)
{
	s_info.get_language_cb = get_language_cb;
}

/**
 * @brief Function which setup startup lanuage for application. It should be called once, after
 * all UI elements were created.
 * @param[in] locale The localization string which is a basis for text language translation.
 */
void view_set_startup_language(char *locale)
{
	_create_current_lang_image(locale);
	_set_language(locale);

	free(locale);
}

/**
 * @brief Function which change lanuage for application.
 * @param[in] locale The localization string which is a basis for text language translation.
 */
void view_change_language(char *locale)
{
	Evas_Object *current = NULL;

	current = _get_lang_item(locale);
	if (!current)
		return;

	strncpy(s_info.new_lang, locale, LOCALE_LEN);
	s_info.lang_changing = true;
	_send_anim_properties(0.5, current);
	free(locale);
}

/**
 * @brief Function adds the provided localization string to the hover selector widget.
 * @param[in] locale The localization string to be added to the hover selector widget.
 */
bool view_add_locale(const char *locale)
{
	if (!locale) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Invalid input argument.");
		return false;
	}

	_create_append_item((void *)locale);

	return true;
}

/**
 * @brief Destroys window and frees its resources.
 */
void view_destroy(void)
{
	if (s_info.win == NULL)
		return;

	evas_object_del(s_info.win);
}

/**
 * @brief Internal callback function invoked when the main window needs to be destroyed.
 * @param[in] data The user data passed to the evas_object_smart_callback_add() function.
 * @param[in] obj The object invoking this callback function.
 * @param[in] event_info The structure containing the information on this event.
 */
static void _delete_win_request_cb(void *data, Evas_Object *obj, void *event_info)
{
	ui_app_exit();
}

/**
 * @brief Internal callback function invoked on HW Back button press.
 * @param[in] data The user data passed to the eext_object_event_callback_add() function.
 * @param[in] obj The object invoking this callback function.
 * @param[in] event_info The structure containing the information on this event.
 */
static void _layout_back_cb(void *data, Evas_Object *obj, void *event_info)
{
	/* Let window go to hide state. */
	elm_win_lower(s_info.win);
}

/**
 * @brief Internal callback function invoked when message from edje arrives.
 * @param[in] data The user data passed to the edje_object_message_handler_set() function.
 * @param[in] obj The object invoking this callback function.
 * @param[in] type Type of the message.
 * @param[in] id Message identifier.
 * @param[in] msg Message structure.
 */
static void _anim_start_end_cb(void *data, Evas_Object *obj, Edje_Message_Type type, int id, void *msg)
{
	Edje_Message_Int *msg_int;

	if (id != MSG_ID_ANIM_IN_PROGRESS)
		return;

	if (type == EDJE_MESSAGE_INT) {
		msg_int = (Edje_Message_Int *)msg;
		s_info.anim_in_progress = msg_int->val;

		if (s_info.anim_in_progress == false)
			elm_scroller_movement_block_set(s_info.scroller, ELM_SCROLLER_MOVEMENT_NO_BLOCK);
		else
			elm_scroller_movement_block_set(s_info.scroller, ELM_SCROLLER_MOVEMENT_BLOCK_HORIZONTAL);
	}
}

/**
 * @brief Callback function invoked when animation of changing language has finished.
 * @param[in] data The user data passed to the callback attachment function.
 * @param[in] obj The layout object emitting the signal.
 * @param[in] emission The name of the signal emitted.
 * @param[in] source The name of the part emitting the signal.
 */
static void _change_language_cb(void *data, Evas_Object *obj, const char *emission, const char *source)
{
	_set_language((const char *)data);
}

/**
 * @brief Callback function invoked, when "change global language" was clicked.
 * @param[in] data The user data passed to the callback attachment function.
 * @param[in] obj The layout object emitting the signal.
 * @param[in] emission The name of the signal emitted.
 * @param[in] source The name of the part emitting the signal.
 */
static void _change_global_lang_cb(void *data, Evas_Object *obj, const char *emission, const char *source)
{
	app_control_h service;

	if (app_control_create(&service) != APP_CONTROL_ERROR_NONE) {
		dlog_print(DLOG_INFO, LOG_TAG, "Failed to create app control handler");
		return;
	}

	app_control_set_operation(service, APP_CONTROL_OPERATION_SETTING);

	if (app_control_send_launch_request(service, NULL, NULL) != APP_CONTROL_ERROR_NONE)
		dlog_print(DLOG_INFO, LOG_TAG, "Failed to launch language settings!");

	app_control_destroy(service);
}

/**
 * @brief Callback function invoked when animation of current item needs to change image.
 * @param[in] data The user data passed to the callback attachment function.
 * @param[in] obj The layout object emitting the signal.
 * @param[in] emission The name of the signal emitted.
 * @param[in] source The name of the part emitting the signal.
 */
static void _change_image_cb(void *data, Evas_Object *obj, const char *emission, const char *source)
{
	_change_current_lang_image(s_info.new_lang);
}

/**
 * @brief Internal callback function invoked on mouse down over an evas object.
 * In this case the object is a flag image. This function blocks scrolling during swipe gesture.
 * @param[in] data Custom data passed to the callback attachment function.
 * @param[in] e The evas surface of the object.
 * @param[in] obj Object to which the handler is attached.
 * @param[in] event_info The information on the event occurred.
 */
static void _mouse_down_cb(void *data, Evas *e, Evas_Object *obj, void *event_info)
{
	elm_scroller_movement_block_set(s_info.scroller, ELM_SCROLLER_MOVEMENT_BLOCK_HORIZONTAL);
}

/**
 * @brief Internal callback function invoked on mouse up over an evas object.
 * In this case the object is a flag image. This function enables scrolling after swipe gesture.
 * @param[in] data Custom data passed to the callback attachment function.
 * @param[in] e The evas surface of the object.
 * @param[in] obj Object to which the handler is attached.
 * @param[in] event_info The information on the event occurred.
 */
static void _mouse_up_cb(void *data, Evas *e, Evas_Object *obj, void *event_info)
{
	elm_scroller_movement_block_set(s_info.scroller, ELM_SCROLLER_MOVEMENT_NO_BLOCK);
}

/**
 * @brief Internal callback function invoked when item is deleted.
 * It frees all the memory allocated and assigned to the item being deleted .
 * @param[in] data Custom data passed to the callback attachment function.
 * @param[in] e The evas surface of the object.
 * @param[in] obj Object to which the handler is attached.
 * @param[in] event_info The information on the event occurred.
 */
static void _item_on_del_cb(void *data, Evas *e, Evas_Object *obj, void *event_info)
{
	if (data)
		free(data);
}

/**
 * @brief Internal callback function which is invoked at start of the gesture
 * performed over the gesture layer.
 * This fuction is attached using elm_gesture_layer_cb_set() function.
 * @param[in] data Custom data passed to the callback attachment function.
 * @param[in] event_info The information on the event occurred.
 */
static Evas_Event_Flags _n_finger_line_start_cb(void *data, void *event_info)
{
	struct timeval tv = {0,};

	gettimeofday(&tv, NULL);
	s_info.drag_start_t = (double)tv.tv_sec + (double)tv.tv_usec / 1000000;

	return EVAS_EVENT_FLAG_NONE;
}

/**
 * @brief Internal callback function which is invoked at end of the gesture.
 * This fuction is attached using elm_gesture_layer_cb_set() function.
 * @param[in] data Custom data passed to the callback attachment function.
 * @param[in] event_info The information on the event occurred.
 */
static Evas_Event_Flags _n_finger_line_end_cb(void *data, void *event_info)
{
	int x = -1;
	int y = -1;
	int w = -1;
	int h = -1;
	int scroll_dir = 0;
	double diff;
	struct timeval tv = {0,};
	Evas_Object *image = NULL;
	Evas_Object *cur_img = NULL;
	Elm_Gesture_Line_Info *line_info = (Elm_Gesture_Line_Info *) event_info;
	Elm_Gesture_Momentum_Info *moment_info = &(line_info->momentum);
	data_pass_t *s_data = (data_pass_t *)data;

	gettimeofday(&tv, NULL);

	s_info.drag_end_t = (double)tv.tv_sec + (double)tv.tv_usec / 1000000;
	diff = s_info.drag_end_t - s_info.drag_start_t;

	image = elm_object_part_content_get(s_data->parent, PART_AVAILABLE_LANGS_ITEM_IMAGE);
	if (!image) {
		dlog_print(DLOG_INFO, LOG_TAG, "Function elm_object_part_content_get() failed!");
		return EVAS_EVENT_FLAG_NONE;
	}

	cur_img = elm_object_part_content_get(s_info.layout_language, PART_LANGUAGE_CUR_LANG);
	if (!cur_img)
		return EVAS_EVENT_FLAG_NONE;

	evas_object_geometry_get(cur_img, &x, &y, &w, &h);
	/* Substract indicator height from real height */
	y = y - INDICATOR_H;

	if (_check_scroll_direction(line_info->angle, &scroll_dir)) {
		int h_page = -1;
		int v_page = -1;
		elm_scroller_current_page_get(s_info.scroller, &h_page, &v_page);
		elm_scroller_page_bring_in(s_info.scroller, h_page + scroll_dir, v_page);
	} else if (_check_swipe_direction(line_info->angle,
									(double)moment_info->x1, (double)moment_info->y1,
									(double)moment_info->x2, (double)moment_info->y2,
									(double)w, (double)(x + w / 2), (double)(y + h / 2)))
		if (!s_info.lang_changing) {
			strncpy(s_info.new_lang, s_data->locale, LOCALE_LEN);
			s_info.lang_changing = true;
			_send_anim_properties(diff, s_data->parent);
		}

	return EVAS_EVENT_FLAG_NONE;
}

/**
 * @brief Internal callback function invoked when the scroller's page is changed.
 * @param[in] data The user data passed to the evas_object_smart_callback_add() function.
 * @param[in] obj The object invoking this callback function.
 * @param[in] event_info The structure containing the information on this event.
 */
static void _content_scrolled_cb(void *data, Evas_Object *obj, void *event_info)
{
	int id_cur = -1;
	int id_old = -1;
	int h_max = -1;

	if (s_info.anim_in_progress == true)
		return;

	elm_scroller_last_page_get(s_info.scroller, &h_max, NULL);

	_toggle_indicator_dot(s_info.current_page, h_max, INDICATOR_OFF, &id_old);
	if (id_old < -1 || id_old > MAX_INDICATOR_PAGE_NO)
		return;

	elm_scroller_current_page_get(s_info.scroller, &s_info.current_page, NULL);

	_toggle_indicator_dot(s_info.current_page, h_max, INDICATOR_ON, &id_cur);
	if (id_cur < -1 || id_cur > MAX_INDICATOR_PAGE_NO)
			return;

	_turn_on_off_dot_indicator(id_old, id_cur);
}

/**
 * @brief Internal function which send a message to EDJE layout with drag time
 * and starts the animation.
 * @param[in] time_diff The duration of drag gesture (in seconds).
 * @param[in] layout The layout object.
 */
static void _send_anim_properties(double time_diff, Evas_Object *layout)
{
	Edje_Message_Float msg = { .val = time_diff };
	Evas_Object *old_lang = _get_lang_item(s_info.current_lang);

	edje_object_message_send(elm_layout_edje_get(s_info.layout_language), EDJE_MESSAGE_FLOAT, MSG_ID_START_ANIM, &msg);
	edje_object_message_send(elm_layout_edje_get(layout), EDJE_MESSAGE_FLOAT, MSG_ID_START_ANIM, &msg);

	if (old_lang)
		edje_object_message_send(elm_layout_edje_get(old_lang), EDJE_MESSAGE_FLOAT, MSG_ID_START_REVERSE_ANIM, &msg);
}

/**
 * @brief Internal function which gets the current language item from the
 * available languages list.
 * @param[in] locale The localization string used to search for associated item.
 * @return This function returns a list's item representing currently selected language.
 * If this function fails, then NULL is returned.
 */
static Evas_Object *_get_lang_item(const char* locale)
{
	Eina_List *items =  elm_box_children_get(s_info.box);
	Eina_List *it = NULL;
	Evas_Object *list_item = NULL;
	const char *data = NULL;

	EINA_LIST_FOREACH(items, it, list_item) {
		if (!list_item)
			continue;

		data = (const char *)evas_object_data_get(list_item, ITEM_INFO);
		if (!data)
			continue;

		if (_is_same_string(data, locale))
			return list_item;
	}

	eina_list_free(items);

	return NULL;
}

/**
 * @brief Internal functions which checks whether a flag image was dragged into
 * the proper direction (part with current flag).
 * It determines line equation using start and end point of gesture, which
 * is then put to the circle equation and determinant is calculated.
 * If the line and the circle has common points, then determinant is be grether than 0.
 * @param[in] angle The linear gesture angle.
 * @param[in] x1 The linear gesture's starting point X.
 * @param[in] y1 The linear gesture's starting point Y.
 * @param[in] x2 The linear gesture's ending point X.
 * @param[in] y2 The linear gesture's ending point Y.
 * @param[in] r The radius of a circle.
 * @param[in] a The circle's center point X.
 * @param[in] b The circle's center point Y.
 */
static bool _check_swipe_direction(double angle, double x1, double y1, double x2, double y2, double r, double a, double b)
{
	double e_a;
	double e_b;
	double e_c;
	double c;
	double d;
	double det;

	/* Check if line is going into right direction */
	if ((angle > 90) && (angle < 270))
		return false;

	/**
	 * Circle equation: (x-a)^2 + (y+b)^2 = r^2
	 * Line eqation: y = -(cx + d)
	 */

	/* Calculate coefficient of line equation */
	c = -1 * ((y2 - y1) / (x2 - x1));
	d = -1 * ((-1 * (y2 - y1) * x1 + (x2 - x1) * y1) / (x2 - x1));

	/**
	 * Calculate coefficient of equation obtained from equate line and circle equations
	 * (c^2 +1)x^2 +(2bc -2a +2cd)x +(a^2 +b^2 +2bd +d^2 -r^2) = 0
	 */

	e_a = ((c * c) + 1);
	e_b = ((2 * b * c) + (-2 * a) + (2 * c * d));
	e_c = ((a * a) + (b * b) + (2 * b * d) + (d * d) - (r * r));

	/**
	 * Calculate determinant of quadratic equation,
	 * if det >= 0 that means line has 2 or 1 common points with circle equation
	 */
	det = e_b * e_b - 4 * e_a * e_c;

	return (det >= 0);
}

/**
 * @brief Internal functions which checks whether a flag image was dragged left
 * or right. This function supports the scroll gesture recognition.
 * @param[in] angle The linear gesture angle.
 * @param[out] direction The gesture direction. For right to left swipe,
 * the 'direction' argument takes the value of 1. For left to right swipe,
 * the 'direction' argument takes the value of -1. For any other directions,
 * the argument takes the value of 0.
 * @return This function returns 'true' if the 'direction' argument is non-zero,
 * otherwise 'false' is returned.
 */
static bool _check_scroll_direction(double angle, int *direction)
{
	if (!direction) {
		dlog_print(DLOG_ERROR, LOG_TAG, "_check_scroll_direction() failed. Invalid input argument.");
		return false;
	}

	*direction = 0;

	if (angle >= SWIPE_RIGHT_ANGLE_LO && angle <= SWIPE_RIGHT_ANGLE_HI)
		*direction = -1;
	else if (angle >= SWIPE_LEFT_ANGLE_LO && angle <= SWIPE_LEFT_ANGLE_HI)
		*direction = 1;

	return (*direction != 0);
}

/**
 * @brief Internal function which applies the provided language to the target text.
 * The plain localization string provided by i18n is not sufficient for the
 * Elementary translation engine. In order to change the language successfully,
 * the coding standard suffix must be added. In this case UTF-8 standard is used,
 * so to the localization string the '.UTF-8' suffix is appended.
 * @param[in] locale The localization string which is a basis for text language translation.
 * @return The function returns 'true' if the language was set successfully,
 * otherwise 'false' is returned.
 */
static bool _set_language(const char *locale)
{
	int str_len;
	char *localization_string;

	if (!locale) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Invalid input arguments.");
		return false;
	}

	s_info.current_lang = locale;
	str_len = strlen(locale) + strlen(".UTF-8") + 1;
	localization_string = (char *)calloc(str_len, sizeof(char));
	snprintf(localization_string, str_len, "%s.UTF-8", locale);

	dlog_print(DLOG_INFO, LOG_TAG, "Language changed to: %s (%s)", locale, localization_string);

	elm_language_set(localization_string);

	free(localization_string);

	_display_text_translation();

	return true;
}

/**
 * @brief Internal function which adds the gesture layer for linear gesture detection.
 * @param[in] parent The parent object for gesture layer.
 */
static Eina_Bool _add_line_gesture_layer(Evas_Object *parent, Evas_Object *layout, const char *locale)
{
	Evas_Object *gest_layer = NULL;
	data_pass_t *s_data = (data_pass_t *)malloc(sizeof(data_pass_t));

	if (!s_data) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Function _add_line_gesture_layer() fails to allocate the memory.");
		return EINA_FALSE;
	}

	s_data->parent = parent;
	s_data->locale = locale;

	evas_object_event_callback_add(parent, EVAS_CALLBACK_DEL, _item_on_del_cb, s_data);

	gest_layer = elm_gesture_layer_add(parent);
	if (!gest_layer) {
		dlog_print(DLOG_INFO, LOG_TAG, "Function elm_gesture_layer_add() failed.");
		free(s_data);
		return EINA_FALSE;
	}

	elm_gesture_layer_attach(gest_layer, parent);

	elm_gesture_layer_cb_set(gest_layer, ELM_GESTURE_N_LINES, ELM_GESTURE_STATE_START, _n_finger_line_start_cb,  NULL);
	elm_gesture_layer_cb_set(gest_layer, ELM_GESTURE_N_LINES, ELM_GESTURE_STATE_END, _n_finger_line_end_cb, s_data);

	evas_object_event_callback_add(layout, EVAS_CALLBACK_MOUSE_DOWN, _mouse_down_cb, NULL);
	evas_object_event_callback_add(layout, EVAS_CALLBACK_MOUSE_UP, _mouse_up_cb, NULL);

	return EINA_TRUE;
}

/**
 * @brief Internal function which creates specfic title for each language.
 * @param[in] locale The base to create the title.
 * @param[out] title The title for each language.
 * @param[in] title_size The size of the ouput title buffer.
 */
static void _create_title(const char *locale, char *title, int title_size)
{
	i18n_uchar result_w[TITLE_LENGTH] = {0,};
	i18n_uchar result_upper[TITLE_LENGTH] = {0,};
	char result_lang[TITLE_LENGTH] = {0,};
	char result_country[TITLE_LENGTH] = {0,};
	i18n_error_code_e error = I18N_ERROR_NONE;

	i18n_ulocale_get_display_language(locale, locale, result_w, TITLE_LENGTH);
	i18n_ustring_to_upper(result_upper, TITLE_LENGTH, result_w, i18n_ustring_get_length(result_w), locale, &error);
	if (error != I18N_ERROR_NONE) {
		dlog_print(DLOG_INFO, LOG_TAG, "Function i18n_ustring_to_upper() failed! Invalid parameters.");
		return;
	}

	i18n_ustring_copy_au(result_lang, result_upper);

	i18n_ulocale_get_display_country(locale, locale, result_w, TITLE_LENGTH);
	i18n_ustring_copy_au(result_country, result_w);

	snprintf(title, 4 * TITLE_LENGTH, "%s<subtitle><br>(%s)</subtitle>", result_lang, result_country);
}

/**
 * @brief Internal function which creates fully qualified path to the provided resource file.
 * @param[in] edj_file_in The file name.
 * @param[out] edj_path_out The fully qualified path to the edj_file_in file.
 */
static void _get_app_resource(const char *edj_file_in, char *edj_path_out)
{
	char *res_path = app_get_resource_path();
	if (res_path) {
		snprintf(edj_path_out, PATH_MAX, "%s%s", res_path, edj_file_in);
		free(res_path);
	}
}

/**
 * @brief Internal function which creates fully qualified path to the flag image.
 * @param[in] locale_in The file name.
 * @param[in] folder Folder name containing flag files.
 * @param[out] flag_path_out The fully qualified path to the edj_file_in file.
 */
static void _get_country_flag_path(const char *locale_in, const char* folder, char *flag_path_out)
{
	char *res_path = app_get_resource_path();
	if (res_path) {
		snprintf(flag_path_out, PATH_MAX, "%simages/%s/%s.png", res_path, folder, locale_in);
		free(res_path);
	}
}

/**
 * @brief Internal function which creates the item's layout.
 * @param[in] locale The country's locale, which is used to create its flag representation.
 * @return The function returns item object, if was created successfully,
 * otherwise 'NULL' is returned.
 */
static Evas_Object *_create_item(const char *locale, Evas_Object *parent)
{
	Evas_Object *layout = NULL;
	Evas_Object *image = NULL;
	char flag_path[PATH_MAX] = {0, };
	char title[256] = {0,};

	layout = view_create_layout(parent, EDJ_AVAILABLE_LANGS, GRP_AVAILABLE_LANGS_ITEM);
	if (!layout)
		return NULL;

	if (locale) {
		_get_country_flag_path(locale, FOLDER_FLAG_LIST, flag_path);

		image = view_create_image(layout, flag_path, PART_AVAILABLE_LANGS_ITEM_IMAGE);
		if (!image)
			return NULL;

		if (!_add_line_gesture_layer(layout, image, locale))
			return NULL;

		if (_is_same_string(s_info.current_lang, locale))
			edje_object_signal_emit(elm_layout_edje_get(layout), SIGNAL_LANG_IS_CHOSEN, "");

		evas_object_data_set(layout, ITEM_INFO, locale);

		elm_layout_signal_callback_add(layout, SIGNAL_ANIM_FINISH, PART_AVAILABLE_LANGS_ITEM_IMAGE, _change_language_cb, (void *)locale);

		_create_title(locale, title, sizeof(title));
		elm_object_part_text_set(layout, PART_AVAILABLE_LANGS_ITEM_TITLE, title);
	}

	evas_object_size_hint_align_set(layout, EVAS_HINT_FILL, EVAS_HINT_FILL);
	evas_object_size_hint_weight_set(layout, EVAS_HINT_EXPAND, EVAS_HINT_EXPAND);

	evas_object_show(layout);

	return layout;
}

/**
 * @brief Internal function which creates and appends an item to the box.
 * @param[in] data The Item data for new item creation.
 * @return The function returns 'true' if the item was created and appended successfully,
 * otherwise 'false' is returned.
 */
static bool _create_append_item(void *data)
{
	Evas_Object *item = _create_item((const char *)data, s_info.box);
	if (!item)
		return false;

	elm_box_pack_end(s_info.box, item);

	return true;
}

/**
 * @brief Internal function which creates an image of current language.
 * @param[in] locale The localization string which is a basis for text language translation.
 * @return The function returns an image object if it was created successfully,
 * otherwise 'NULL' is returned.
 */
static Evas_Object *_create_current_lang_image(const char *locale)
{
	Evas_Object *image = NULL;
	char flag_path[PATH_MAX] = {0, };

	if (locale) {
		_get_country_flag_path(locale, FOLDER_FLAG_CURRENT, flag_path);
		image = view_create_image(s_info.layout_language, flag_path, PART_LANGUAGE_CUR_LANG);
	}

	return image;
}

/**
 * @brief Internal function which changes an image of current language.
 * @param[in] locale The localization string which is a basis for text language translation.
 * @return The function returns 'EINA_TRUE' if the image was changed successfully,
 * otherwise 'EINA_FALSE' is returned.
 */
static bool _change_current_lang_image(const char *locale)
{
	Evas_Object *image = NULL;
	char flag_path[PATH_MAX] = {0, };

	_get_country_flag_path(locale, FOLDER_FLAG_CURRENT, flag_path);

	image = elm_object_part_content_get(s_info.layout_language, PART_LANGUAGE_CUR_LANG);
	if (!image) {
		dlog_print(DLOG_INFO, LOG_TAG, "Function elm_object_part_content_get() failed!");
		return false;
	}

	if (!elm_image_file_set(image, flag_path, NULL)) {
		dlog_print(DLOG_INFO, LOG_TAG, "Function elm_image_file_set() failed!");
		return false;
	}

	evas_object_show(image);

	return true;
}

/**
 * @brief Internal function which turns on or off page indicator by sending a message to the edje script.
 * @param[in] id_old The index of page which indicator is to be turned off.
 * @param[in] id_cur The index of page which indicator is to be turned on.
 * @param[in] i_s indicator state (on or off).
 */
static void _turn_on_off_dot_indicator(int id_old, int id_cur)
{
	Edje_Message_Int_Set *msg = malloc(sizeof(Edje_Message_Int_Set) + sizeof(int));
	if (!msg) {
		dlog_print(DLOG_INFO, LOG_TAG, "Function _turn_on_off_dot_indicator() failed to allocate memory.");
		return;
	}

	msg->count = 2;
	msg->val[0] = id_old;
	msg->val[1] = id_cur;

	edje_object_message_send(elm_layout_edje_get(s_info.layout_available_langs), EDJE_MESSAGE_INT_SET, MSG_ID_TOGGLE_IND, msg);
	free(msg);
}

/**
 * @brief Internal function which turns on/off the page indicator's dot relevant
 * to the currently visible page.
 * @param[in] page_number The page number.
 * @param[in] max_page The max page number.
 * @param[in] i_s indicator state (on or off).
 */
static void _toggle_indicator_dot(int page_number, int max_page, indicator_state i_s, int *id)
{
	float items_per_indicator = (max_page - AVAILABLE_LANGS_LIST_ITEMS_PER_PAGE + 1) / NUMBER_OF_PAGE_INDICATORS;
	*id = page_number / items_per_indicator;
	if (*id > MAX_INDICATOR_PAGE_NO)
		*id = *id - 1;
}

/**
 * @brief Internal function which creates the main layout of the application.
 * @return The function returns 'EINA_TRUE' if the main layout was created successfully,
 * otherwise 'EINA_FALSE' is returned.
 */
static Eina_Bool _create_layout_main(void)
{
	s_info.conform = view_create_conformant(s_info.win);
	if (!s_info.conform)
		return EINA_FALSE;

	s_info.layout_main = view_create_set_layout(s_info.conform, EDJ_MAIN, GRP_MAIN, NULL);
	if (!s_info.layout_main)
		return EINA_FALSE;

	eext_object_event_callback_add(s_info.layout_main, EEXT_CALLBACK_BACK, _layout_back_cb, NULL);

	return EINA_TRUE;
}

/**
 * @brief Internal function which creates the language selector layout.
 * @return The function returns 'EINA_TRUE' if the language selector layout was created successfully,
 * otherwise 'EINA_FALSE' is returned.
 */
static Eina_Bool _create_layout_language(void)
{
	s_info.layout_language = view_create_set_layout(s_info.layout_main, EDJ_LANGUAGE, GRP_LANGUAGE, PART_MAIN_LANGUAGE);
	if (!s_info.layout_language)
		return EINA_FALSE;

	elm_layout_signal_callback_add(s_info.layout_language, SIGNAL_LANG_IMAGE_CHANGE, PART_LANGUAGE_CUR_LANG, _change_image_cb, NULL);

	return EINA_TRUE;
}

/**
 * @brief Internal function which creates the container to hold all the supported languages.
 * @return The function returns 'EINA_TRUE' if the container was created successfully,
 * otherwise 'EINA_FALSE' is returned.
 */
static Eina_Bool _create_layout_available_languages(void)
{
	s_info.layout_available_langs = view_create_set_layout(s_info.layout_main, EDJ_AVAILABLE_LANGS, GRP_AVAILABLE_LANGS, PART_MAIN_AVAILABLE_LANGS);
	if (!s_info.layout_available_langs)
		return EINA_FALSE;

	s_info.scroller =  view_create_scroller(s_info.layout_available_langs, PART_AVAILABLE_LANGS_LIST);
	if (!s_info.scroller)
		return EINA_FALSE;

	evas_object_smart_callback_add(s_info.scroller, "scroll,anim,stop", _content_scrolled_cb, NULL);

	s_info.box = view_create_box(s_info.scroller);
	if (!s_info.box)
		return EINA_FALSE;

	edje_object_signal_emit(elm_layout_edje_get(s_info.layout_available_langs), SIGNAL_IND_0_ON, "");
	edje_object_message_handler_set(elm_layout_edje_get(s_info.layout_available_langs), _anim_start_end_cb, NULL);

	return EINA_TRUE;
}

/**
 * @brief Internal function which creates the button selector layout.
 * @return The function returns 'EINA_TRUE' if the main layout was created successfully,
 * otherwise 'EINA_FALSE' is returned.
 */
static Eina_Bool _create_layout_button(void)
{
	Evas_Object *button = view_create_set_layout(s_info.layout_main, EDJ_BUTTON, GRP_BUTTON, PART_MAIN_BUTTON);
	if (!button)
		return EINA_FALSE;

	elm_layout_signal_callback_add(button, SIGNAL_BUTTON_CLICKED, PART_BUTTON_BUTTON, _change_global_lang_cb, NULL);

	return EINA_TRUE;
}

/**
 * @brief Internal function which displays the translated text.
 * Internally, the predefined text tags are used as an input arguments for
 * i18n. These tags are defined in translation files (res/po).
 */
static void _display_text_translation(void)
{
	elm_object_part_text_set(s_info.layout_language, PART_LANGUAGE_MESSAGE, i18n_get_text("IDS_BODY_HELLO_WORLD"));
	s_info.lang_changing = false;
}

/**
 * @brief Internal function which checks whether two strings are the same.
 * @param[in] s1 The first string.
 * @param[in] s2 The second string.
 * @return This function returns 'true' is both given strings are the same,
 * otherwise 'false' is returned.
 */
static bool _is_same_string(const char *s1, const char *s2)
{
	if (!s1 || !s2)
		return false;

	return (strncmp(s1, s2, strlen(s1)) == 0 && strlen(s1) == strlen(s2));
}
